import shutil
import subprocess
import time
import click
import getpass
import json
import requests
# import web3
import socket
import os
# from dotenv import load_dotenv
import yaml
# from kubernetes import client,config
from .oasees_ml_ops import sdk_manager_manifest
from .ipfs_client import ipfs_get,ipfs_upload
import platform
from pathlib import Path
import sys


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if not ctx.invoked_subcommand:
        bash_script = r'''
        Y='\x1b[38;2;252;220;132m'
        G='\x1b[38;2;60;148;140m'
        O='\x1b[38;2;244;172;92m'
        N='\033[0m' # No Color
        B='\x1b[38;2;4;188;252m'

        echo -e "                ${Y}.......           ${O}=======${N}                "             
        echo -e "               ${Y}.........        ${O}==========${N}               "            
        echo -e "               ${Y}.......... ${O}================${N}               "             
        echo -e "               ${Y}.........${O}==================${N}               "             
        echo -e "                ${Y}....... ${O}========= ======${N}                 "               
        echo -e "             ${Y}.......     ${O}=======     ${Y}.......${N}             "           
        echo -e "     ${O}=====  ${Y}.........${O}=========${Y}...............  .....${N}     "
        echo -e "   ${O}=========${Y}.........${O}======   ${Y}........................${N}   "
        echo -e "   ${O}==========${Y}.......  ${O}===== ${Y}........ .................   "
        echo -e "   ${O}==========  ${Y}...  ${O}====   ${Y}....  ${O}==== ===   ${Y}..........   "
        echo -e "   ${O} ========  ${Y}......${O}====  ${B}${G}**${B}  ${G}**${N} ${O}=========   ${Y}........    "
        echo -e "   ${O}   ======= ${Y}......     ${B}${G}**${B}****${G}**${N}    ${O}===== =====${Y}...      "   
        echo -e "   ${O}    ======== ${Y}.. ..  ${G}**${B}*******${G}**${N}   ${Y}..${O}== ========       "    
        echo -e "   ${O}    ========   ${Y}....${G}**${B}*********${G}**${Y}..... ${O}=========       "    
        echo -e "   ${O}    ============${Y}..  ${G}**${B}*******${G}**${N}   ${Y}.....${O}========       "    
        echo -e "   ${O}   ${Y}...${O}==== ======     ${G}**${B}****${G}**${O}====${Y}......${O}=========${N}     "  
        echo -e "   ${Y}.........  ${O}========== ${G} **  ** ${N} ${O}====${Y}......  ${O}=========   "
        echo -e "   ${Y}..........   .${O}=   ===  ${Y}....  ${O}====   ${Y}...  ${O}==========   "
        echo -e "   ${Y}................. ......... ${O}======${Y}........${O}=========   "
        echo -e "   ${Y}........................    ${O}=====${Y}.........${O}========    "
        echo -e "     ${Y}.....  ..............${O}========   ${Y}........${O}   ===      "   
        echo -e "             ${Y}.......     ${O}=========    ${Y}......             "
        echo -e "                   ${O}===============${Y}.......                "
        echo -e "                  ${O}===============${Y}.........               "
        echo -e "                 ${O}==========     ${Y}..........               " 
        echo -e "                 ${O}==========      ${Y}.........               "
        echo -e "                  ${O}========        ${Y}.......                "
                                                
        echo -e ${B}"  ___    _    ____  _____ _____ ____    ____  ____  _  __ "
        echo -e ${B}" / _ \  / \  / ___|| ____| ____/ ___|  / ___||  _ \| |/ / "
        echo -e ${B}"| | | |/ _ \ \___ \|  _| |  _| \___ \  \___ \| | | | ' /  "
        echo -e ${B}"| |_| / ___ \ ___) | |___| |___ ___) |  ___) | |_| | . \  "
        echo -e ${B}" \___/_/   \_\____/|_____|_____|____/  |____/|____/|_|\_\ "${N}
        '''

        # Execute the Bash script and capture its output
        result = subprocess.run(['bash', '-c', bash_script], capture_output=True, text=True)
        click.echo(result.stdout)
        cli(['--help'])


# @click.option('--price', required=False, type=float, default=0, help="")
def kompose_install():
    '''Installs kompose on the user's path'''

    # Detect system architecture
    machine = platform.machine().lower()
    
    # Determine download URL based on architecture
    if machine in ('x86_64', 'amd64'):
        url = "https://github.com/kubernetes/kompose/releases/download/v1.35.0/kompose-linux-amd64"
    elif machine in ('arm64', 'aarch64'):
        url = "https://github.com/kubernetes/kompose/releases/download/v1.35.0/kompose-linux-arm64"
    else:
        print(f"Unsupported architecture: {machine}")
        sys.exit(1)
    
    print(f"Detected architecture: {machine}")
    print(f"Downloading Kompose from: {url}")
    
    try:
        # Download Kompose
        subprocess.run(["curl", "-L", "-o", "kompose", url], check=True)
        
        # Make it executable
        subprocess.run(["chmod", "+x", "kompose"], check=True)
        
        # Move to /usr/local/bin (requires sudo)
        subprocess.run(["sudo", "mv", "kompose", "/usr/local/bin/"], check=True)
        
        click.secho("Kompose installed successfully", fg="green")
        return True
    except subprocess.CalledProcessError as e:
        click.secho(f"Error during installation: {e}",fg="red")
        sys.exit(1)
def helm_install():
    '''Installs helm on the user's path'''

    try:
        print("Installing Helm...")
        subprocess.run([
            "curl", "-fsSL", "-o", "get_helm.sh",
            "https://raw.githubusercontent.com/helm/helm/main/scripts/get-helm-3"
        ], check=True)
        subprocess.run(["chmod", "700", "get_helm.sh"], check=True)
        subprocess.run(["./get_helm.sh"], check=True)
        Path("get_helm.sh").unlink()  # Cleanup
        
        click.secho("Helm installed successfully", fg="green")
        return True
    except subprocess.CalledProcessError as e:
        click.secho(f"Failed to install Helm: {e}", fg="red")
        sys.exit(1)

def helm_add_oasees_repo():
    '''Adds and / or updates the repository containing all of OASEES' charts'''

    add_cmd = ["repo", "add", "oasees-charts", "https://oasees.github.io/helm-charts"]
    update_cmd = ["repo", "update", "oasees-charts"]
    
    click.echo("*Adding OASEES helm chart repository...*")
    click.echo("------------------------------------")
    run_helm_command(add_cmd)

    click.echo('\n')
    click.echo("*Updating OASEES chart repository...*")
    click.echo("------------------------------------")
    run_helm_command(update_cmd)


def helm_install_oasees_user():
    cmd = ["upgrade", "--install", "oasees-user", "oasees-charts/oasees-user", "--create-namespace"]
    run_helm_command(cmd)


def run_helm_command(cmd:list):
    
    try:
        result = subprocess.run(
            ["helm"] + cmd,
            check=True,
            text=True
        )

    except subprocess.CalledProcessError as e:
        return e.stderr


def check_required_tools():
    """Check if helm and kompose are installed and available in PATH."""
    helm_installed = shutil.which("helm") is not None
    kompose_installed = shutil.which("kompose") is not None
    
    click.echo("Checking if the required tools are installed on your platform:")
    click.echo(f"Helm: {'✅ installed' if helm_installed else '❌ not found'}")
    click.echo(f"Kompose: {'✅ installed' if kompose_installed else '❌ not found'}")

    if helm_installed and kompose_installed:
        return True
    else:
        if click.confirm("\nHelm and Kompose are needed in order to use all of the OASEES SDK features. Would you like to install them now?"):
            helm_install()
            kompose_install()
            return True
        else:
            return False

def get_nodes():
    '''Retrieves information on the cluster's nodes.'''
    try:
        result = subprocess.run(["kubectl", "get", "nodes", "-o", "jsonpath='{.items[*].metadata}'"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        node_names = result.stdout.strip().strip("'").split()
        return node_names
    except FileNotFoundError:
        click.echo("Error: No connection to cluster found.")

@cli.command()
@click.option('--expose-ip', required=False, help="Expose a node's IP if you plan ")
def init():

    if(check_required_tools()):
        try:
            curl = subprocess.Popen(['curl','-sfL', 'https://get.k3s.io'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            result = subprocess.check_output(['sh','-s','-','--write-kubeconfig-mode','644', '--write-kubeconfig', '/home/'+getpass.getuser()+'/.kube/config', '--node-label', 'user='+getpass.getuser()], stdin=curl.stdout)
            click.echo(result)
            curl.wait()

        except subprocess.CalledProcessError as e:
            return e.stderr

        helm_add_oasees_repo()
        click.echo('\n')
        helm_install_oasees_user()
        click.echo('\n')

        click.secho("Kubernetes cluster initialized successfully!",fg="green")
    else:
        click.secho("Please install Helm and / or Kompose before trying again.", fg="red")

@cli.command()
@click.argument('role', type=str)
def uninstall(role):
    '''Runs the appropriate k3s uninstallation script based on the role provided.'''
    try:
        if(role=='master'):
            result = subprocess.run(['/usr/local/bin/k3s-uninstall.sh'],  stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            click.echo("Cluster uninstalled successfully.")
        elif (role=='agent'):
            result = subprocess.run(['/usr/local/bin/k3s-agent-uninstall.sh'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            click.echo("Agent uninstalled successfully.")
        else:
            click.echo("Enter a valid role (master / agent)")
    except FileNotFoundError:
        click.echo("Error: Uninstall executable not found.")


@cli.command()
def get_token():
    '''Retrieves the token required to join the cluster.'''

    try:
        token = subprocess.run(['sudo', 'cat','/var/lib/rancher/k3s/server/token'],  stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        click.echo(token.stdout)
    except FileNotFoundError:
        click.echo("Error: No token found.")


@cli.command()
@click.option('--ip', required=True, help="The cluster's master ip address.")
@click.option('--token', required=True, help="The cluster's master token.")
@click.option('--iface', required=False, help="The network interface you want to join with (tun0 if you're using a VPN connection).")
def join(ip,token,iface):
    '''Joins the current machine to the specified cluster, using the specified interface if one is provided.'''
    try:
        curl = subprocess.Popen(['curl','-sfL', 'https://get.k3s.io'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        if(iface):
            result = subprocess.check_output(['sh','-s','-','--flannel-iface', iface, '--node-label', 'user='+getpass.getuser()], env={'K3S_URL' : 'https://'+ip+':6443', 'K3S_TOKEN': token}, stdin=curl.stdout)
        else:
            result = subprocess.check_output(['sh','-s','-','--node-label', 'user='+getpass.getuser()], env={'K3S_URL' : 'https://'+ip+':6443', 'K3S_TOKEN': token}, stdin=curl.stdout)
        click.echo(result)
        curl.wait()
    except FileNotFoundError:
        click.echo("Error: K3S cluster could not be joined.\n")


@cli.command()
@click.argument('compose_file', type=str)
@click.argument('output_dir', type=str)
def convert_app(compose_file, output_dir):
    """Convert Docker Compose to Kubernetes manifests using kompose"""
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir)
    
    print("Converting Docker Compose using Kompose...")
    subprocess.run(["kompose", "convert", "-f", compose_file, "-o", output_dir], check=True)

    process_yaml_files(output_dir)

def process_yaml_files(output_dir):
    """Process generated YAML files to add nodeSelectors and other modifications"""
    yaml_files = list(Path(output_dir).glob("*.yaml")) + list(Path(output_dir).glob("*.yml"))
    
    for yaml_file in yaml_files:
        print(f"Processing {yaml_file}...")
        
        with open(yaml_file, 'r') as f:
            data = list(yaml.safe_load_all(f))
        
        modified = False
        
        for doc in data:
            if not doc:
                continue
                
            kind = doc.get('kind', '')
            
            # Process Deployments
            if kind == 'Deployment':
                metadata = doc.get('metadata', {})
                annotations = metadata.get('annotations', {})
                
                # Handle nodeSelector
                node_selector_value = annotations.get('oasees.device')
                if node_selector_value:
                    if 'spec' not in doc:
                        doc['spec'] = {}
                    if 'template' not in doc['spec']:
                        doc['spec']['template'] = {}
                    if 'spec' not in doc['spec']['template']:
                        doc['spec']['template']['spec'] = {}
                    
                    doc['spec']['template']['spec']['nodeSelector'] = {
                        'kubernetes.io/hostname': node_selector_value
                    }
                    modified = True
                
                # Handle sensor resources
                sensor_value = annotations.get('oasees.sensor')
                if sensor_value:
                    if 'spec' not in doc:
                        doc['spec'] = {}
                    if 'template' not in doc['spec']:
                        doc['spec']['template'] = {}
                    if 'spec' not in doc['spec']['template']:
                        doc['spec']['template']['spec'] = {}
                    if 'containers' not in doc['spec']['template']['spec']:
                        continue
                    
                    resource_map = {
                        'sound': {'smarter-devices/snd': 1},
                        'video': {'smarter-devices/video0': 1},
                        'bluetooth': {'smarter-devices/hci0': 1}
                    }
                    
                    if sensor_value in resource_map:
                        doc['spec']['template']['spec']['containers'][0]['resources'] = {
                            'limits': resource_map[sensor_value]
                        }
                        modified = True
            
            # Process Services
            elif kind == 'Service':
                expose_value = doc.get('metadata', {}).get('annotations', {}).get('oasees.expose')
                if expose_value:
                    doc['spec']['type'] = 'NodePort'
                    doc['spec']['ports'][0]['nodePort'] = int(expose_value)
                    modified = True
        
        if modified:
            with open(yaml_file, 'w') as f:
                yaml.dump_all(data, f, default_flow_style=False)
@cli.command()
@click.argument('folder', type=str)
def deploy_app(folder):
    """Deploy all YAML/JSON manifests in a folder using kubectl"""
    files = list(Path(folder).glob('*.yaml')) + list(Path(folder).glob('*.yml'))
    
    if not files:
        print(f"No manifests found in {folder}")
        return False

    success = True
    for f in files:
        try:
            print(f"Deploying {f.name}...")
            subprocess.run(['kubectl', 'apply', '-f', str(f)], check=True)
        except subprocess.CalledProcessError:
            print(f"Failed to deploy {f.name}")
            success = False
    
    if success:
        click.secho("All manifests deployed successfully!", fg="green")
    else:
        click.secho("Some manifests failed to deploy.", fg="red")

@cli.command()
@click.argument('folder', type=str)
def delete_app(folder):
    """Delete all YAML/JSON manifests in a folder using kubectl"""
    files = list(Path(folder).glob('*.yaml')) + list(Path(folder).glob('*.yml'))
    
    if not files:
        print(f"No manifests found in {folder}")
        return False

    success = True
    for f in files:
        try:
            print(f"Deploying {f.name}...")
            subprocess.run(['kubectl', 'delete', '-f', str(f)], check=True)
        except subprocess.CalledProcessError:
            print(f"Failed to deploy {f.name}")
            success = False
    
    if success:
        click.secho("All manifests deleted successfully!", fg="green")
    else:
        click.secho("Some manifests failed to delete.", fg="red")

@cli.command()
def enable_gpu_operator():
    '''Enables the GPU Operator on the cluster'''

    click.secho("Enabling GPU Operator...", fg="yellow")

    command = [
        "helm", "install", "gpu-operator", "-n", "gpu-operator", "--create-namespace", "nvidia/gpu-operator",
        "--version=v25.3.0",
        "--set", "toolkit.env[0].name=CONTAINERD_CONFIG",
        "--set", "toolkit.env[0].value=/var/lib/rancher/k3s/agent/etc/containerd/config.toml",
        "--set", "toolkit.env[1].name=CONTAINERD_SOCKET",
        "--set", "toolkit.env[1].value=/run/k3s/containerd/containerd.sock",
        "--set", "toolkit.env[2].name=CONTAINERD_RUNTIME_CLASS",
        "--set", "toolkit.env[2].value=nvidia",
        "--set", "toolkit.env[3].name=CONTAINERD_SET_AS_DEFAULT",
        "--set-string", "toolkit.env[3].value=true"
    ]

    try:
        result = subprocess.run(command, check=True, text=True)
        click.secho("GPU Operator enabled successfully!", fg="green")
    except subprocess.CalledProcessError as e:
        click.secho(f"Failed to enable GPU Operator: {e}", fg="red")
        sys.exit(1)

# @cli.command()
# def mlops_add_node():
#     '''Install the components needed to execute ML workloads on the specified node.'''
#     node_name = ''
#     mount_path = ''
#     gpu_enabled = False

#     nodes = get_nodes()
#     click.echo("Available nodes:")
#     for node in nodes:
#         click.echo(dict(node))

#     click.echo('\n')

#     click.echo("Choose the node you want to use for training:")